/*
 * Copyright (c) 2016 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include "ubjs_selfemptying_list_prv.h"
#include "ubjs_common_prv.h"

ubjs_result ubjs_selfemptying_list_new(ubjs_library *lib, ubjs_list_free_f free_f,
    ubjs_selfemptying_list_callback callback, void *userdata, ubjs_selfemptying_list **pthis)
{
    ubjs_selfemptying_list *this = 0;
    this=(ubjs_selfemptying_list *)(lib->alloc_f)(sizeof(struct ubjs_selfemptying_list));
    this->lib=lib;

    this->list = 0;
    ubjs_list_new(lib, free_f, &(this->list));
    this->callback=callback;
    this->is_in_callback=UFALSE;
    this->userdata=userdata;

    *pthis=this;
    return UR_OK;
}

ubjs_result ubjs_selfemptying_list_free(ubjs_selfemptying_list **pthis)
{
    ubjs_selfemptying_list *this=*pthis;

    ubjs_list_free(&(this->list));
    (this->lib->free_f)(this);
    *pthis=0;
    return UR_OK;
}

ubjs_result ubjs_selfemptying_list_add(ubjs_selfemptying_list *this, void *obj)
{
    unsigned int len = 0;
    void *lobj = 0;

    ubjs_list_add(this->list, obj);

    if (UTRUE == this->is_in_callback)
    {
        return UR_OK;
    }

    this->is_in_callback=UTRUE;
    ubjs_list_len(this->list, &len);
    while (0 < len)
    {
        ubjs_list_remove_first_and_get(this->list, &lobj);
        (this->callback)(this, lobj);
        (this->list->free_f)(lobj);
        ubjs_list_len(this->list, &len);
    }
    this->is_in_callback=UFALSE;
    return UR_OK;
}
